var bsv_lib = nbpay.bsv;
var coin_lib;
async function vbox_init() {
    coin_lib = await coinfly.create('bsv');
}
vbox_init();

const api = (function(W) {

    const bitId_Protocol = "14kxqYv3emHGwf8m6YgSYLQkGCvn9Qrgr9";
    const refund_Protocol = "refund";

    const PLACEHOLDER_LEN = 150;
    const placeholder = " ".repeat(PLACEHOLDER_LEN);

    function b64DecodeUnicode(str) {
        return decodeURIComponent(atob(str).split('').map(function(c) {
            return '%' + ('00' + c.charCodeAt(0).toString(16)).slice(-2);
        }).join(''));
    }
  
    function stringLength(str) {
        var len = 0;
        for (var i=0; i<str.length; i++) {
          var c = str.charCodeAt(i);
          if ((c >= 0x0001 && c <= 0x007e) || (0xff60<=c && c<=0xff9f)) {
            len++;
          }else {
            len+=2;
          }
        }
        return len;
    }

    function backGetFee(id, err, res, fee) {
        if (err) {
            if (typeof(err) == 'object') {
                chrome.send('getFeeCallback', ['FAIL', id.toString(), '0', '0', JSON.stringify(err)]);
            }
            else if (typeof(err) == 'string') {
                if (err.indexOf("No available utxos") != -1
                            || err.indexOf("Insufficient input utxo") != -1) {
                    chrome.send('getFeeCallback', ['NO UTXOS', id.toString(), '0', '0', err.toString()]);
                }
                else{
                    chrome.send('getFeeCallback', ['FAIL', id.toString(), '0', '0', err.toString()]);
                }
            }
            else {
                chrome.send('getFeeCallback', ['FAIL', id.toString(), '0', '0', 'other error']);
            }
        } else {
            chrome.send('getFeeCallback', ['SUCESS', id.toString(), res.inputAmount.toString(), res.getFee().toString(), '']);
        }
    }

    function backSignPay(id, err, res, fee) {
        if (err) {
            if (typeof(err) == 'object') {
                chrome.send('signPayCallback', ['FAIL', id.toString(), '', JSON.stringify(err), '0']);
            }
            else if (typeof(err) == 'string') {
                if (err.indexOf("No available utxos") != -1
                            || err.indexOf("Insufficient input utxo") != -1) {
                    chrome.send('signPayCallback', ['NO UTXOS', id.toString(), '', '', '0']);
                }
                else{
                    chrome.send('signPayCallback', ['FAIL', id.toString(), '', err.toString(), '0']);
                }
            }
            else {
                chrome.send('signPayCallback', ['FAIL', id.toString(), '', 'other error', '0']);
            }
        } else {
            chrome.send('signPayCallback', ['SUCESS', id.toString(), res.toString(), res.hash.toString(), res.getFee().toString()]);
        }
    }

    async function backSend(id, err, res, fee) {
        if (err) {
            if (typeof(err) == 'object') {
                chrome.send('sendPayCallback', ['FAIL', id.toString(), '', JSON.stringify(err), '0']);
            }
            else if (typeof(err) == 'string') {
                if (err.indexOf("No available utxos") != -1
                            || err.indexOf("Insufficient input utxo") != -1) {
                    chrome.send('sendPayCallback', ['NO UTXOS', id.toString(), '', '', '0']);
                }
                else{
                    chrome.send('sendPayCallback', ['FAIL', id.toString(), '', err.toString(), '0']);
                }
            }
            else {
                chrome.send('sendPayCallback', ['FAIL', id.toString(), '', 'other error', '0']);
            }
        } else {
            try {
                var send_op = { tx : res.toString() };
                var send_result = await coin_lib.send(send_op);
                if (send_result.code == 0)
                      chrome.send('sendPayCallback', ['SUCESS', id.toString(), res.toString(), send_result.txid.toString(), res.getFee().toString()]);
                  else
                      chrome.send('sendPayCallback', ['FAIL', id.toString(), res.toString(), send_err.message.toString(), res.getFee().toString()]);
            }
            catch(err){
                chrome.send('sendPayCallback', ['FAIL', id.toString(), res.toString(), err.message.toString(), res.getFee().toString()]);
            }
        }
    }

    function back(type, id, err, res, fee) {
        if (type == 'signpay') {
            backSignPay(id, err, res, fee);
        } else if (type == 'getfee') {
            backGetFee(id, err, res, fee);
        } else if (type == 'sendpay') {
            backSend(id, err, res, fee);
        }
    }

    function genDataToSign(tx) {
        let data2sign = "";
        //console.log(tx);
        tx.inputs.forEach((inp)=>{
            data2sign += inp.prevTxId.toString("hex") + inp.outputIndex;
        }
        );
        tx.outputs.forEach((out)=>{
            const sc = new bsv_lib.Script.fromBuffer(out._scriptBuffer);
            const sc_len = sc.chunks.length;
            if (sc.chunks[1].opcodenum == 106) {
                if (sc.chunks[2].buf.toString() == bitId_Protocol) {
                    console.log("found bitID. PublicKey=" + sc.chunks[3].buf.toString());
                    data2sign += sc.chunks[3].buf.toString() + out._satoshis;//public key
                } else {
                    data2sign += sc.toHex() + out._satoshis;
                }
            } else {
                data2sign += sc.toHex() + out._satoshis;
            }
        }
        );
        return data2sign;
    }

    function getSignature(tx, index, satoshis, script, key, type) {
        const privKey = new bitcoin.PrivateKey(key);
        const sigtype =
            type | bsv_lib.crypto.Signature.SIGHASH_FORKID;
        const flags =
            bsv_lib.Script.Interpreter.SCRIPT_VERIFY_MINIMALDATA |
            bsv_lib.Script.Interpreter.SCRIPT_ENABLE_SIGHASH_FORKID |
            bsv_lib.Script.Interpreter.SCRIPT_ENABLE_MAGNETIC_OPCODES |
            bsv_lib.Script.Interpreter.SCRIPT_ENABLE_MONOLITH_OPCODES;
	      const signature = bsv_lib.Transaction.sighash.sign(
            tx,
            privKey,
            sigtype,
            index,
            script,
            new bsv_lib.crypto.BN(satoshis),
            flags
        );
        const strSignature = signature.toBuffer();
        return bsv_lib.deps.Buffer.concat([
            strSignature,
            bsv_lib.deps.Buffer.from([
                (sigtype || type) & 0xff,
            ]),
        ])
    }

    function base64ToBuffer(str) {
        return new Promise(function (res, rej) {
            var binary_string = b64DecodeUnicode(str);
            var blob = new Blob([binary_string], {
              type: 'text/plain'
            });

            var reader = new FileReader();
            reader.readAsArrayBuffer(blob);
            reader.onload = function (e) {
              res(reader.result);
            }
        });
    }

    async function changeb64Data(type, id, config){
        if(config.pay.inputs && config.pay.inputs instanceof Array){
            let arr = [];
            for(let x = 0; x<config.pay.inputs.length; x++) {
                if(config.pay.inputs[x].unlockingScript){
                    for(let i = 0; i<config.pay.inputs[x].unlockingScript.length; i++) {
                        if(typeof(config.pay.inputs[x].unlockingScript[i]) == 'string' && config.pay.inputs[x].unlockingScript[i].substring(0, 5) === '_b64:') {
                            let promise = new Promise(function (res, rej) {
                                var binary_string = b64DecodeUnicode(config.pay.inputs[x].unlockingScript[i].substring(5));
                                var blob = new Blob([binary_string], {
                                  type: 'text/plain'
                                });

                                var reader = new FileReader();
                                reader.readAsArrayBuffer(blob);
                                reader.onload = function (e) {
                                  res(reader.result);
                                }
                            });
                            arr.push(promise);
                        }
                    }
                }
            }
            Promise.all(arr).then(function (values) {
                //values
                let j = 0;
                for(let x = 0; x<config.pay.inputs.length; x++) {
                    if(config.pay.inputs[x].unlockingScript) {
                        for(let i = 0; i<config.pay.inputs[x].unlockingScript.length; i++) {
                            if(typeof(config.pay.inputs[x].unlockingScript[i]) == 'string' && config.pay.inputs[x].unlockingScript[i].substring(0, 5) === '_b64:') {
                                config.pay.inputs[x].unlockingScript.splice(i, 1, values[j]);
                                j++;
                            }
                        }
                    }
                }
                genTx(type, id, config);
            })
        }
        else{
           genTx(type, id, config);
        }
    }

    function signNBinput(tx, index, satoshis, script, key, unScripts) {
        const privKey = new bsv_lib.PrivateKey(key);
        const pubKey = privKey.publicKey;
        let in_script = new bsv_lib.Script();
        unScripts.forEach((unScript)=>{
            if(unScript == "{OP_PUBKEY}") {
                in_script.add(new bsv_lib.PublicKey(pubKey).toBuffer())
            }
            else if(unScript == "{OP_PUBKEYHASH}") {
                let public_hash = bsv_lib.crypto.Hash.sha256ripemd160(new bsv_lib.PublicKey(pubKey).toBuffer());
                in_script.add(public_hash);
            }
            else if(unScript == "{OP_ADDRESS}") {
                in_script.add(privKey.toAddress().toBuffer())
            }
            else if(unScript == "{OP_ADDRESSHASH}") {
                in_script.add(privKey.toAddress().hashBuffer)
            }
            else if(unScript == "{OP_SIG_HASHALL}") {
                in_script.add(
                    getSignature(tx, index, satoshis, script, key, bsv_lib.crypto.Signature.SIGHASH_ALL)
                );
            }
            else if(unScript == "{OP_SIG_HASHNONE}") {
                in_script.add(
                    getSignature(tx, index, satoshis, script, key, bsv_lib.crypto.Signature.SIGHASH_NONE)
                );
            }
            else if(unScript == "{OP_SIG_HASHSINGLE}") {
                in_script.add(
                    getSignature(tx, index, satoshis, script, key, bsv_lib.crypto.Signature.SIGHASH_SINGLE)
                );
            }
            else if(unScript == "{OP_SIG_HASHANY}") {
                in_script.add(
                    getSignature(tx, index, satoshis, script, key, bsv_lib.crypto.Signature.SIGHASH_ANYONECANPAY)
                );
            }
            else if(typeof(unScript) == 'string') {
                let script = bsv_lib.Script.fromHex(unScript);
                in_script.add(script);
            }
        }
        );
        return in_script;
    }

    async function genTx(type, id, config) {
        var is_have_bitidentity = false;
        config.pay.to.forEach((out)=>{
            if (out.protocol) {
                if(out.protocol == "bitIdentity") {
                    is_have_bitidentity = true;
                    var pubKey;
                    if (out.value.privateKey) {
                        const pk = bsv_lib.PrivateKey.fromWIF(out.value.privateKey);
                        pubKey = bsv_lib.PublicKey.fromPrivateKey(pk);
                    } else if (out.value.publicKey) {
                        pubKey = bsv_lib.PublicKey.fromString(out.value.publicKey);
                    }
                    out["data"] = [bitId_Protocol, pubKey.toString(), placeholder];
                    out["pvalue"] = out.value;
                    out.value = 0;
                }
                else if (out.protocol == "refund") {
                    const ppk = bsv_lib.PrivateKey.fromWIF(config.pay.key);
                    var addressKey = bsv_lib.Address.fromPrivateKey(ppk);
                    out["data"] = [refund_Protocol, addressKey.toString()];
                    out.value = 0;
                }
            }
        }
        );


        if(config.pay.inputs && config.pay.inputs instanceof Array){
            config.pay.inputs.forEach( async (input)=>{
                if (input.txid && typeof(input.index ) == 'number' && typeof(input.value) == 'number' && input.lockingScript && input.privateKey) {
                    input["required"] = true;
                    input["outputIndex"] = input.index;
                    input["script"] = input.lockingScript;

                    delete input.index;
                    delete input.lockingScript;

                    var unlockingScript = input.unlockingScript;
                    delete input.unlockingScript;
                    if (unlockingScript) {
                        input["unlockingScript"] = function (tx, index, satoshis, script, key) {
                            return signNBinput(tx, index, satoshis, script, key, unlockingScript);
                        }
                    }
                }
            }
            );
        }

        console.log(config.pay.to);
        console.log(config.pay.inputs);
        var tx_result = await coin_lib.build(config);
        if (is_have_bitidentity === false || tx_result.code != 0) {
            back(type, id, tx_result.err, tx_result.tx, tx_result.fee);
        } else {
            if (tx_result.tx) {
                console.log("rawtx1=", tx_result.tx.toString());
                let data2sign = genDataToSign(tx_result.tx);
                console.log("data2sign=" + data2sign);
                var hash = bsv_lib.crypto.Hash.sha256(bsv_lib.deps.Buffer.from(data2sign));
                console.log("hash=" + hash.toString("hex"));
                config.pay.to.forEach((out)=>{
                    if (out.protocol && out.protocol == "bitIdentity") {
                        if (out.pvalue.privateKey) {
                            const pKey = bsv_lib.PrivateKey.fromWIF(out.pvalue.privateKey);
                            var sig = bsv_lib.crypto.ECDSA.sign(hash, pKey);
                            var sig_str = sig.toString();
                            out.data[2] = sig_str.concat(" ".repeat(PLACEHOLDER_LEN - sig_str.length));
                            console.log("signature=" + sig_str);
                        }
                        delete out.protocol;
                        delete out.pvalue;
                    }
                }
                );
                console.log(config.pay.to);

                var tx_result_again = await coin_lib.build(config);
                if (tx_result_again.tx) {
                    console.log("data2sign111=" + genDataToSign(tx_result_again.tx));
                    console.log("rawtx=" + tx_result_again.tx.toString());
                }
                back(type, id, tx_result_again.err, tx_result_again.tx, tx_result_again.fee);
            } else {
                back(type, id, tx_result.err, tx_result.tx, tx_result.fee);
            }
        }
    }

    async function sign(args) {
        if (args.length == 3) {
            var id = args[0];
            var hash = bsv_lib.deps.Buffer.from(args[2], 'hex');
            if (hash.byteLength == 32) {
                var sig = await coin_lib.sign(args[1], hash);
                chrome.send('signCallback', ['SUCESS', id.toString(), sig.toString()]);
            } else {
                chrome.send('signCallback', ['FAIL', id.toString(), '']);
            }
        }
    }

    async function createKey(args) {
        if (args.length == 2) {
            console.log(args);
            var id = args[0];
            var time = args[1]
            let privateKey = await coin_lib.createPrivateKey();
            let publicKey = await coin_lib.getPublicKey(privateKey);
            let address = await coin_lib.getAddress(publicKey)
            chrome.send('createKeyCallback', [id.toString(), time.toString(), privateKey.toString(), publicKey.toString(), address.toString()]);
        }
    }

    async function getPubAdrKey(args) {
        if (args.length == 2) {
            console.log(args);
            var id = args[0];
            var privateKey = args[1];
            try {
              for(var i = 0; i<privateKey.length; i++){
                let publicKey = await coin_lib.getPublicKey(privateKey[i].privateKey);
                let address = await coin_lib.getAddress(publicKey)
                privateKey[i]["publicKey"] = publicKey.toString();
                privateKey[i]["address"] = address.toString();
              }

              var json_str = JSON.stringify(privateKey);
              chrome.send('getPubAdrKeyCallback', [id.toString(), json_str]);
            }
            catch(err){
                chrome.send('getPubAdrKeyCallback', [id.toString(), '']);
            }
        }
    }

   async function signPay(args) {
        if (args.length == 2) {
            var id = args[0];
            const content = args[1];
            var content_obj = JSON.parse(content);
            changeb64Data('signpay', id, content_obj);
        }
    }

   async function sendPay(args) {
        if (args.length == 2) {
            var id = args[0];
            const content = args[1];
            var content_obj = JSON.parse(content);
            changeb64Data('sendpay', id, content_obj);
        }
    }

   async function getFee(args) {
        if (args.length == 2) {
            var id = args[0];
            const content = args[1];
            var content_obj = JSON.parse(content);
            changeb64Data('getfee', id, content_obj);
        }
    }

    async function updateNBDomain(args) {
        if (args.length == 5) {
            var id = args[0];
            const nbdomain = args[1];
            const address = args[2];
            const priv = args[3];
            const content = args[4];
            var content_obj = JSON.parse(content);
            try {
            await nblib.init({
	            API:"https://tnode.nbdomain.com/api/",
	            token:"02a43685fc7613626164d36555f7bfee2adafa4f4d35d2816860c61aca15505c58",
	            filepayKey:"44h9cKf4VHUvdpbRnG8KER1qCwx3oEjqho7TFBZv23BFgMtewE7k4kXPJbfv1EPQsi",
	            debug:true,
	            sendByNode:true,
	            enable_write:true
            });

            const domain = await nblib.getDomain(nbdomain);
	          if(!domain){
		          return;
	          }
            domain.setPrivateKey(priv);

            const res = await domain.updateKey(content_obj);
			console.log("updateNBDomain", res);
            chrome.send('updateNBDomainCallback', ['SUCESS', id.toString(), JSON.stringify(res)]);
            }
            catch(err){
                chrome.send('updateNBDomainCallback', ['FAIL', id.toString(), err.message]);
            }

        }
    }

    return {
        sign,
        createKey,
        getPubAdrKey,
        signPay,
        sendPay,
        getFee,
        updateNBDomain,
    }

}
)(window);
